import pytest
import fiona
from fiona.drvsupport import supported_drivers, _driver_supports_mode
from fiona.errors import DriverError
from .conftest import driver_extensions
from fiona.env import GDALVersion


def test_bounds_point():
    g = {'type': 'Point', 'coordinates': [10, 10]}
    assert fiona.bounds(g) == (10, 10, 10, 10)


def test_bounds_line():
    g = {'type': 'LineString', 'coordinates': [[0, 0], [10, 10]]}
    assert fiona.bounds(g) == (0, 0, 10, 10)


def test_bounds_polygon():
    g = {'type': 'Polygon', 'coordinates': [[[0, 0], [10, 10], [10, 0]]]}
    assert fiona.bounds(g) == (0, 0, 10, 10)


def test_bounds_z():
    g = {'type': 'Point', 'coordinates': [10, 10, 10]}
    assert fiona.bounds(g) == (10, 10, 10, 10)


ignore_write_drivers = set(['CSV', 'GPX', 'GPSTrackMaker', 'DXF', 'DGN', 'MapInfo File'])
write_drivers = [driver for driver, raw in supported_drivers.items() if
                 _driver_supports_mode(driver, 'w') and driver not in ignore_write_drivers]


@pytest.mark.parametrize('driver', write_drivers)
def test_bounds(tmpdir, driver):
    """Test if bounds are correctly calculated after writing

    """

    if driver == 'BNA' and GDALVersion.runtime() < GDALVersion(2, 0):
        # BNA driver segfaults with gdal 1.11
        return

    extension = driver_extensions.get(driver, "bar")
    path = str(tmpdir.join('foo.{}'.format(extension)))

    with fiona.open(path, 'w',
                    driver=driver,
                    schema={'geometry': 'Point',
                            'properties': [('title', 'str')]},
                    fiona_force_driver=True) as c:

        c.writerecords([{'geometry': {'type': 'Point', 'coordinates': (1.0, 10.0)},
                         'properties': {'title': 'One'}}])

        try:
            bounds = c.bounds
            assert bounds == (1.0, 10.0, 1.0, 10.0)
        except Exception as e:
            assert isinstance(e, DriverError)

        c.writerecords([{'geometry': {'type': 'Point', 'coordinates': (2.0, 20.0)},
                         'properties': {'title': 'Two'}}])

        try:
            bounds = c.bounds
            assert bounds == (1.0, 10.0, 2.0, 20.0)
        except Exception as e:
            assert isinstance(e, DriverError)



"""Implementation of Apache VFS schemes and URLs."""

import os
import sys
import re
from fiona.compat import urlparse


# Supported URI schemes and their mapping to GDAL's VSI suffix.
# TODO: extend for other cloud plaforms.
SCHEMES = {
    'ftp': 'curl',
    'gzip': 'gzip',
    'http': 'curl',
    'https': 'curl',
    's3': 's3',
    'tar': 'tar',
    'zip': 'zip',
    'gs': 'gs',
}

CURLSCHEMES = set([k for k, v in SCHEMES.items() if v == 'curl'])

# TODO: extend for other cloud plaforms.
REMOTESCHEMES = set([k for k, v in SCHEMES.items() if v in ('curl', 's3', 'gs')])


def valid_vsi(vsi):
    """Ensures all parts of our vsi path are valid schemes."""
    return all(p in SCHEMES for p in vsi.split('+'))

def is_remote(scheme):
    if scheme is None:
        return False
    return any(p in REMOTESCHEMES for p in scheme.split('+'))


def vsi_path(path, vsi=None, archive=None):
    # If a VSI and archive file are specified, we convert the path to
    # an OGR VSI path (see cpl_vsi.h).
    if vsi:
        prefix = '/'.join('vsi{0}'.format(SCHEMES[p]) for p in vsi.split('+'))
        if archive:
            result = '/{0}/{1}{2}'.format(prefix, archive, path)
        else:
            result = '/{0}/{1}'.format(prefix, path)
    else:
        result = path

    return result


def parse_paths(uri, vfs=None):
    """Parse a URI or Apache VFS URL into its parts

    Returns: tuple
        (path, scheme, archive)
    """
    archive = scheme = None
    path = uri
    # Windows drive letters (e.g. "C:\") confuse `urlparse` as they look like
    # URL schemes
    if sys.platform == "win32" and re.match("^[a-zA-Z]\\:", path):
        return path, None, None
    if vfs:
        parts = urlparse(vfs)
        scheme = parts.scheme
        archive = parts.path
        if parts.netloc and parts.netloc != 'localhost':
            archive = parts.netloc + archive
    else:
        parts = urlparse(path)
        scheme = parts.scheme
        path = parts.path
        if parts.netloc and parts.netloc != 'localhost':
            if scheme.split("+")[-1] in CURLSCHEMES:
                # We need to deal with cases such as zip+https://server.com/data.zip
                path = "{}://{}{}".format(scheme.split("+")[-1], parts.netloc, path)
            else:
                path = parts.netloc + path
        if scheme in SCHEMES:
            parts = path.split('!')
            path = parts.pop() if parts else None
            archive = parts.pop() if parts else None

    scheme = None if not scheme else scheme
    return path, scheme, archive



"""Tests for geometry objects."""

import pytest

from fiona._geometry import (GeomBuilder, geometryRT)
from fiona.errors import UnsupportedGeometryTypeError


def geometry_wkb(wkb):
    try:
        wkb = bytes.fromhex(wkb)
    except AttributeError:
        wkb = wkb.decode('hex')
    return GeomBuilder().build_wkb(wkb)


def test_ogr_builder_exceptions():
    geom = {'type': "Bogus", 'coordinates': None}
    with pytest.raises(ValueError):
        geometryRT(geom)


@pytest.mark.parametrize('geom_type, coordinates', [
    ('Point', (0.0, 0.0)),
    ('LineString', [(0.0, 0.0), (1.0, 1.0)]),
    ('Polygon',
     [[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]]),
    ('MultiPoint', [(0.0, 0.0), (1.0, 1.0)]),
    ('MultiLineString', [[(0.0, 0.0), (1.0, 1.0)]]),
    ('MultiPolygon',
     [[[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]]]),
])
def test_round_tripping(geom_type, coordinates):
    result = geometryRT({'type': geom_type, 'coordinates': coordinates})
    assert result['type'] == geom_type
    assert result['coordinates'] == coordinates


@pytest.mark.parametrize('geom_type, coordinates', [
    ('Polygon', [[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)]]),
    ('MultiPolygon', [[[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)]]]),
])
def test_implicitly_closed_round_tripping(geom_type, coordinates):
    result = geometryRT({'type': geom_type, 'coordinates': coordinates})
    assert result['type'] == geom_type
    result_coordinates = result['coordinates']
    while not isinstance(coordinates[0], tuple):
        result_coordinates = result_coordinates[0]
        coordinates = coordinates[0]
    assert result_coordinates[:-1] == coordinates


def test_geometry_collection_round_trip():
    geom = {
        'type': "GeometryCollection",
        'geometries': [
            {'type': "Point", 'coordinates': (0.0, 0.0)}, {
                'type': "LineString",
                'coordinates': [(0.0, 0.0), (1.0, 1.0)]}]}

    result = geometryRT(geom)
    assert len(result['geometries']) == 2
    assert [g['type'] for g in result['geometries']] == ['Point', 'LineString']


def test_point_wkb():
    # Hex-encoded Point (0 0)
    wkb = "010100000000000000000000000000000000000000"
    geom = geometry_wkb(wkb)
    assert geom['type'] == "Point"
    assert geom['coordinates'] == (0.0, 0.0)


def test_line_wkb():
    # Hex-encoded LineString (0 0, 1 1)
    wkb = ("01020000000200000000000000000000000000000000000000000000000000f03f"
           "000000000000f03f")
    geom = geometry_wkb(wkb)
    assert geom['type'] == "LineString"
    assert geom['coordinates'] == [(0.0, 0.0), (1.0, 1.0)]


def test_polygon_wkb():
    # 1 x 1 box (0, 0, 1, 1)
    wkb = ("01030000000100000005000000000000000000f03f000000000000000000000000"
           "0000f03f000000000000f03f0000000000000000000000000000f03f0000000000"
           "0000000000000000000000000000000000f03f0000000000000000")
    geom = geometry_wkb(wkb)
    assert geom['type'], "Polygon"
    assert len(geom['coordinates']) == 1
    assert len(geom['coordinates'][0]) == 5
    x, y = zip(*geom['coordinates'][0])
    assert min(x) == 0.0
    assert min(y) == 0.0
    assert max(x) == 1.0
    assert max(y) == 1.0


def test_multipoint_wkb():
    wkb = ("010400000002000000010100000000000000000000000000000000000000010100"
           "0000000000000000f03f000000000000f03f")
    geom = geometry_wkb(wkb)
    assert geom['type'] == "MultiPoint"
    assert geom['coordinates'] == [(0.0, 0.0), (1.0, 1.0)]


def test_multilinestring_wkb():
    # Hex-encoded LineString (0 0, 1 1)
    wkb = ("010500000001000000010200000002000000000000000000000000000000000000"
           "00000000000000f03f000000000000f03f")
    geom = geometry_wkb(wkb)
    assert geom['type'] == "MultiLineString"
    assert len(geom['coordinates']) == 1
    assert len(geom['coordinates'][0]) == 2
    assert geom['coordinates'][0] == [(0.0, 0.0), (1.0, 1.0)]


def test_multipolygon_wkb():
    # [1 x 1 box (0, 0, 1, 1)]
    wkb = ("01060000000100000001030000000100000005000000000000000000f03f000000"
           "0000000000000000000000f03f000000000000f03f000000000000000000000000"
           "0000f03f00000000000000000000000000000000000000000000f03f0000000000"
           "000000")
    geom = geometry_wkb(wkb)
    assert geom['type'] == "MultiPolygon"
    assert len(geom['coordinates']) == 1
    assert len(geom['coordinates'][0]) == 1
    assert len(geom['coordinates'][0][0]) == 5
    x, y = zip(*geom['coordinates'][0][0])
    assert min(x) == 0.0
    assert min(y) == 0.0
    assert max(x) == 1.0
    assert max(y) == 1.0



"""$ fio collect"""


from functools import partial, wraps
import json
import logging

import click
import cligj

from fiona.fio import helpers, options, with_context_env
from fiona.transform import transform_geom


@click.command(short_help="Collect a sequence of features.")
@cligj.precision_opt
@cligj.indent_opt
@cligj.compact_opt
@click.option('--record-buffered/--no-record-buffered', default=False,
              help="Economical buffering of writes at record, not collection "
              "(default), level.")
@click.option('--ignore-errors/--no-ignore-errors', default=False,
              help="log errors but do not stop serialization.")
@options.src_crs_opt
@click.option('--with-ld-context/--without-ld-context', default=False,
              help="add a JSON-LD context to JSON output.")
@click.option('--add-ld-context-item', multiple=True,
              help="map a term to a URI and add it to the output's JSON LD "
                   "context.")
@click.option('--parse/--no-parse', default=True,
              help="load and dump the geojson feature (default is True)")
@click.pass_context
@with_context_env
def collect(ctx, precision, indent, compact, record_buffered, ignore_errors,
            src_crs, with_ld_context, add_ld_context_item, parse):
    """Make a GeoJSON feature collection from a sequence of GeoJSON
    features and print it."""
    logger = logging.getLogger(__name__)
    stdin = click.get_text_stream('stdin')
    sink = click.get_text_stream('stdout')

    dump_kwds = {'sort_keys': True}
    if indent:
        dump_kwds['indent'] = indent
    if compact:
        dump_kwds['separators'] = (',', ':')
    item_sep = compact and ',' or ', '

    if src_crs:
        if not parse:
            raise click.UsageError("Can't specify --src-crs with --no-parse")
        transformer = partial(transform_geom, src_crs, 'EPSG:4326',
                              antimeridian_cutting=True, precision=precision)
    else:
        def transformer(x):
            return x

    first_line = next(stdin)

    # If parsing geojson
    if parse:
        # If input is RS-delimited JSON sequence.
        if first_line.startswith(u'\x1e'):
            def feature_text_gen():
                buffer = first_line.strip(u'\x1e')
                for line in stdin:
                    if line.startswith(u'\x1e'):
                        if buffer:
                            feat = json.loads(buffer)
                            feat['geometry'] = transformer(feat['geometry'])
                            yield json.dumps(feat, **dump_kwds)
                        buffer = line.strip(u'\x1e')
                    else:
                        buffer += line
                else:
                    feat = json.loads(buffer)
                    feat['geometry'] = transformer(feat['geometry'])
                    yield json.dumps(feat, **dump_kwds)
        else:
            def feature_text_gen():
                feat = json.loads(first_line)
                feat['geometry'] = transformer(feat['geometry'])
                yield json.dumps(feat, **dump_kwds)

                for line in stdin:
                    feat = json.loads(line)
                    feat['geometry'] = transformer(feat['geometry'])
                    yield json.dumps(feat, **dump_kwds)

    # If *not* parsing geojson
    else:
        # If input is RS-delimited JSON sequence.
        if first_line.startswith(u'\x1e'):
            def feature_text_gen():
                buffer = first_line.strip(u'\x1e')
                for line in stdin:
                    if line.startswith(u'\x1e'):
                        if buffer:
                            yield buffer
                        buffer = line.strip(u'\x1e')
                    else:
                        buffer += line
                else:
                    yield buffer
        else:
            def feature_text_gen():
                yield first_line
                for line in stdin:
                    yield line

    try:
        source = feature_text_gen()

        if record_buffered:
            # Buffer GeoJSON data at the feature level for smaller
            # memory footprint.
            indented = bool(indent)
            rec_indent = "\n" + " " * (2 * (indent or 0))

            collection = {
                'type': 'FeatureCollection',
                'features': []}
            if with_ld_context:
                collection['@context'] = helpers.make_ld_context(
                    add_ld_context_item)

            head, tail = json.dumps(collection, **dump_kwds).split('[]')

            sink.write(head)
            sink.write("[")

            # Try the first record.
            try:
                i, first = 0, next(source)
                if with_ld_context:
                    first = helpers.id_record(first)
                if indented:
                    sink.write(rec_indent)
                sink.write(first.replace("\n", rec_indent))
            except StopIteration:
                pass
            except Exception as exc:
                # Ignoring errors is *not* the default.
                if ignore_errors:
                    logger.error(
                        "failed to serialize file record %d (%s), "
                        "continuing",
                        i, exc)
                else:
                    # Log error and close up the GeoJSON, leaving it
                    # more or less valid no matter what happens above.
                    logger.critical(
                        "failed to serialize file record %d (%s), "
                        "quiting",
                        i, exc)
                    sink.write("]")
                    sink.write(tail)
                    if indented:
                        sink.write("\n")
                    raise

            # Because trailing commas aren't valid in JSON arrays
            # we'll write the item separator before each of the
            # remaining features.
            for i, rec in enumerate(source, 1):
                try:
                    if with_ld_context:
                        rec = helpers.id_record(rec)
                    if indented:
                        sink.write(rec_indent)
                    sink.write(item_sep)
                    sink.write(rec.replace("\n", rec_indent))
                except Exception as exc:
                    if ignore_errors:
                        logger.error(
                            "failed to serialize file record %d (%s), "
                            "continuing",
                            i, exc)
                    else:
                        logger.critical(
                            "failed to serialize file record %d (%s), "
                            "quiting",
                            i, exc)
                        sink.write("]")
                        sink.write(tail)
                        if indented:
                            sink.write("\n")
                        raise

            # Close up the GeoJSON after writing all features.
            sink.write("]")
            sink.write(tail)
            if indented:
                sink.write("\n")

        else:
            # Buffer GeoJSON data at the collection level. The default.
            collection = {
                'type': 'FeatureCollection',
                'features': []}
            if with_ld_context:
                collection['@context'] = helpers.make_ld_context(
                    add_ld_context_item)

            head, tail = json.dumps(collection, **dump_kwds).split('[]')
            sink.write(head)
            sink.write("[")
            sink.write(",".join(source))
            sink.write("]")
            sink.write(tail)
            sink.write("\n")

    except Exception:
        logger.exception("Exception caught during processing")
        raise click.Abort()



"""pytest fixtures and automatic test data generation."""

import copy
import json
import os.path
import shutil
import tarfile
import zipfile
from collections import OrderedDict
from click.testing import CliRunner
import pytest

import fiona
from fiona.crs import from_epsg
from fiona.env import GDALVersion

driver_extensions = {'DXF': 'dxf',
                     'CSV': 'csv',
                     'ESRI Shapefile': 'shp',
                     'FileGDB': 'gdb',
                     'OpenFileGDB': 'gdb',
                     'GML': 'gml',
                     'GPX': 'gpx',
                     'GPSTrackMaker': 'gtm',
                     'MapInfo File': 'tab',
                     'DGN': 'dgn',
                     'GPKG': 'gpkg',
                     'GeoJSON': 'json',
                     'GeoJSONSeq': 'geojsons',
                     'GMT': 'gmt',
                     'OGR_GMT': 'gmt',
                     'BNA': 'bna',
                     'FlatGeobuf': 'fgb'}


def pytest_collection_modifyitems(config, items):

    # Fiona contains some tests that depend only on GDALs behavior.
    # E.g. some test the driver specific access modes maintained in
    # fiona/drvsupport.py for different GDAL versions.
    # These tests can fail on exotic architectures (e.g. not returning 
    # the exact same value)
    # We explicitly enable these tests on Fiona CI using pytest -m gdal
    # and hide these tests otherwise.
    markers_options = config.getoption("-m", "")
    if "gdal" not in markers_options:
        skip_gdal = pytest.mark.skip(reason="use '-m gdal' to run GDAL related tests.")
        for item in items:
            gdal_marker = item.get_closest_marker("gdal")
            if gdal_marker is not None and gdal_marker.name == "gdal":
                item.add_marker(skip_gdal)


def pytest_report_header(config):
    headers = []
    # gdal version number
    gdal_release_name = fiona.get_gdal_release_name()
    headers.append('GDAL: {} ({})'.format(gdal_release_name, fiona.get_gdal_version_num()))
    supported_drivers = ", ".join(sorted(list(fiona.drvsupport.supported_drivers.keys())))
    # supported drivers
    headers.append("Supported drivers: {}".format(supported_drivers))
    return '\n'.join(headers)


def get_temp_filename(driver):

    basename = "foo"
    extension = driver_extensions.get(driver, "bar")
    prefix = ""
    if driver == 'GeoJSONSeq':
        prefix = "GeoJSONSeq:"

    return "{prefix}{basename}.{extension}".format(prefix=prefix,
                                                   basename=basename,
                                                   extension=extension)


_COUTWILDRNP_FILES = [
    'coutwildrnp.shp', 'coutwildrnp.shx', 'coutwildrnp.dbf', 'coutwildrnp.prj']


def _read_file(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()


has_gpkg = "GPKG" in fiona.drvsupport.supported_drivers.keys()
has_gpkg_reason = "Requires geopackage driver"
requires_gpkg = pytest.mark.skipif(not has_gpkg, reason=has_gpkg_reason)


@pytest.fixture(scope='function')
def gdalenv(request):
    import fiona.env

    def fin():
        if fiona.env.local._env:
            fiona.env.delenv()
            fiona.env.local._env = None
    request.addfinalizer(fin)


@pytest.fixture(scope='session')
def data_dir():
    """Absolute file path to the directory containing test datasets."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))


@pytest.fixture(scope='function')
def data(tmpdir, data_dir):
    """A temporary directory containing a copy of the files in data."""
    for filename in _COUTWILDRNP_FILES:
        shutil.copy(os.path.join(data_dir, filename), str(tmpdir))
    return tmpdir


@pytest.fixture(scope='session')
def path_curves_line_csv(data_dir):
    """Path to ```curves_line.csv``"""
    return os.path.join(data_dir, 'curves_line.csv')


@pytest.fixture(scope='session')
def path_test_tin_shp(data_dir):
    """Path to ```test_tin.shp``"""
    return os.path.join(data_dir, 'test_tin.shp')

@pytest.fixture(scope='session')
def path_test_tin_csv(data_dir):
    """Path to ```test_tin.csv``"""
    return os.path.join(data_dir, 'test_tin.csv')

@pytest.fixture(scope='session')
def path_coutwildrnp_shp(data_dir):
    """Path to ```coutwildrnp.shp``"""
    return os.path.join(data_dir, 'coutwildrnp.shp')


@pytest.fixture(scope='session')
def path_coutwildrnp_zip(data_dir):
    """Creates ``coutwildrnp.zip`` if it does not exist and returns the absolute
    file path."""
    path = os.path.join(data_dir, 'coutwildrnp.zip')
    if not os.path.exists(path):
        with zipfile.ZipFile(path, 'w') as zip:
            for filename in _COUTWILDRNP_FILES:
                zip.write(os.path.join(data_dir, filename), filename)
    return path


@pytest.fixture(scope='session')
def path_grenada_geojson(data_dir):
    """Path to ```grenada.geojson```"""
    return os.path.join(data_dir, 'grenada.geojson')


@pytest.fixture(scope='session')
def bytes_coutwildrnp_zip(path_coutwildrnp_zip):
    """The zip file's bytes"""
    with open(path_coutwildrnp_zip, 'rb') as src:
        return src.read()


@pytest.fixture(scope='session')
def path_coutwildrnp_tar(data_dir):
    """Creates ``coutwildrnp.tar`` if it does not exist and returns the absolute
    file path."""
    path = os.path.join(data_dir, 'coutwildrnp.tar')
    if not os.path.exists(path):
        with tarfile.open(path, 'w') as tar:
            for filename in _COUTWILDRNP_FILES:
                tar.add(
                    os.path.join(data_dir, filename),
                    arcname=os.path.join('testing', filename))
    return path


@pytest.fixture(scope='session')
def path_coutwildrnp_json(data_dir):
    """Creates ``coutwildrnp.json`` if it does not exist and returns the absolute
    file path."""
    path = os.path.join(data_dir, 'coutwildrnp.json')
    if not os.path.exists(path):
        name = _COUTWILDRNP_FILES[0]
        with fiona.open(os.path.join(data_dir, name), 'r') as source:
            features = [feat for feat in source]
        my_layer = {
            'type': 'FeatureCollection',
            'features': features}
        with open(path, 'w') as f:
            f.write(json.dumps(my_layer))
    return path


@pytest.fixture(scope='session')
def bytes_grenada_geojson(path_grenada_geojson):
    """The geojson as bytes."""
    with open(path_grenada_geojson, 'rb') as src:
        return src.read()


@pytest.fixture(scope='session')
def path_coutwildrnp_gpkg(data_dir):
    """Creates ``coutwildrnp.gpkg`` if it does not exist and returns the absolute
    file path."""
    if not has_gpkg:
        raise RuntimeError("GDAL has not been compiled with GPKG support")
    path = os.path.join(data_dir, 'coutwildrnp.gpkg')
    if not os.path.exists(path):
        filename_shp = _COUTWILDRNP_FILES[0]
        path_shp = os.path.join(data_dir, filename_shp)
        with fiona.open(path_shp, "r") as src:
            meta = copy.deepcopy(src.meta)
            meta["driver"] = "GPKG"
            with fiona.open(path, "w", **meta) as dst:
                dst.writerecords(src)
    return path


@pytest.fixture(scope='session')
def path_gpx(data_dir):
    return os.path.join(data_dir, 'test_gpx.gpx')


@pytest.fixture(scope='session')
def feature_collection():
    """GeoJSON feature collection on a single line."""
    return _read_file(os.path.join('data', 'collection.txt'))


@pytest.fixture(scope='session')
def feature_collection_pp():
    """Same as above but with pretty-print styling applied."""
    return _read_file(os.path.join('data', 'collection-pp.txt'))


@pytest.fixture(scope='session')
def feature_seq():
    """One feature per line."""
    return _read_file(os.path.join('data', 'sequence.txt'))


@pytest.fixture(scope='session')
def feature_seq_pp_rs():
    """Same as above but each feature has pretty-print styling"""
    return _read_file(os.path.join('data', 'sequence-pp.txt'))


@pytest.fixture(scope='session')
def runner():
    """Returns a ```click.testing.CliRunner()`` instance."""
    return CliRunner()


@pytest.fixture(scope='class')
def uttc_path_coutwildrnp_zip(path_coutwildrnp_zip, request):
    """Make the ``path_coutwildrnp_zip`` fixture work with a
    ``unittest.TestCase()``.  ``uttc`` stands for unittest test case."""
    request.cls.path_coutwildrnp_zip = path_coutwildrnp_zip


@pytest.fixture(scope='class')
def uttc_path_coutwildrnp_tar(path_coutwildrnp_tar, request):
    """Make the ``path_coutwildrnp_tar`` fixture work with a
    ``unittest.TestCase()``.  ``uttc`` stands for unittest test case."""
    request.cls.path_coutwildrnp_tar = path_coutwildrnp_tar


@pytest.fixture(scope='class')
def uttc_path_coutwildrnp_json(path_coutwildrnp_json, request):
    """Make the ``path_coutwildrnp_json`` fixture work with a
    ``unittest.TestCase()``.  ``uttc`` stands for unittest test case."""
    request.cls.path_coutwildrnp_json = path_coutwildrnp_json


@pytest.fixture(scope='class')
def uttc_data_dir(data_dir, request):
    """Make the ``data_dir`` fixture work with a ``unittest.TestCase()``.
    ``uttc`` stands for unittest test case."""
    request.cls.data_dir = data_dir


@pytest.fixture(scope='class')
def uttc_path_gpx(path_gpx, request):
    """Make the ``path_gpx`` fixture work with a ``unittest.TestCase()``.
    ``uttc`` stands for unittest test case."""
    request.cls.path_gpx = path_gpx


# GDAL 2.3.x silently converts ESRI WKT to OGC WKT
# The regular expression below will match against either
WGS84PATTERN = 'GEOGCS\["(?:GCS_WGS_1984|WGS 84)",DATUM\["WGS_1984",SPHEROID\["WGS[_ ]84"'

# Define helpers to skip tests based on GDAL version
gdal_version = GDALVersion.runtime()

requires_only_gdal1 = pytest.mark.skipif(
    gdal_version.major != 1,
    reason="Only relevant for GDAL 1.x")

requires_gdal2 = pytest.mark.skipif(
    not gdal_version.major >= 2,
    reason="Requires GDAL 2.x")

requires_gdal21 = pytest.mark.skipif(
    not gdal_version.at_least('2.1'),
    reason="Requires GDAL 2.1.x")

requires_gdal22 = pytest.mark.skipif(
    not gdal_version.at_least('2.2'),
    reason="Requires GDAL 2.2.x")

requires_gdal24 = pytest.mark.skipif(
    not gdal_version.at_least('2.4'),
    reason="Requires GDAL 2.4.x")

requires_gdal_lt_3 = pytest.mark.skipif(
    not gdal_version.major < 3,
    reason="Requires GDAL < 3")

requires_gdal3 = pytest.mark.skipif(
    not gdal_version.major >= 3,
    reason="Requires GDAL 3.x")

travis_only = pytest.mark.skipif(
    not os.getenv("TRAVIS", "false") == "true",
    reason="Requires travis CI environment"
)


@pytest.fixture(scope="class")
def unittest_data_dir(data_dir, request):
    """Makes data_dir available to unittest tests"""
    request.cls.data_dir = data_dir


@pytest.fixture(scope="class")
def unittest_path_coutwildrnp_shp(path_coutwildrnp_shp, request):
    """Makes shapefile path available to unittest tests"""
    request.cls.path_coutwildrnp_shp = path_coutwildrnp_shp


@pytest.fixture()
def testdata_generator():
    """ Helper function to create test data sets for ideally all supported drivers
    """

    def get_schema(driver):
        special_schemas = {'CSV': {'geometry': None, 'properties': OrderedDict([('position', 'int')])},
                           'BNA': {'geometry': 'Point', 'properties': {}},
                           'DXF': {'properties': OrderedDict(
                               [('Layer', 'str'),
                                ('SubClasses', 'str'),
                                ('Linetype', 'str'),
                                ('EntityHandle', 'str'),
                                ('Text', 'str')]),
                               'geometry': 'Point'},
                           'GPX': {'geometry': 'Point',
                                   'properties': OrderedDict([('ele', 'float'), ('time', 'datetime')])},
                           'GPSTrackMaker': {'properties': OrderedDict([]), 'geometry': 'Point'},
                           'DGN': {'properties': OrderedDict([]), 'geometry': 'LineString'},
                           'MapInfo File': {'geometry': 'Point', 'properties': OrderedDict([('position', 'str')])}
                           }

        return special_schemas.get(driver, {'geometry': 'Point', 'properties': OrderedDict([('position', 'int')])})

    def get_crs(driver):
        special_crs = {'MapInfo File': from_epsg(4326)}
        return special_crs.get(driver, None)

    def get_records(driver, range):
        special_records1 = {'CSV': [{'geometry': None, 'properties': {'position': i}} for i in range],
                            'BNA': [{'geometry': {'type': 'Point', 'coordinates': (0.0, float(i))}, 'properties': {}}
                                    for i
                                    in range],
                            'DXF': [
                                {'geometry': {'type': 'Point', 'coordinates': (0.0, float(i))},
                                 'properties': OrderedDict(
                                     [('Layer', '0'),
                                      ('SubClasses', 'AcDbEntity:AcDbPoint'),
                                      ('Linetype', None),
                                      ('EntityHandle', str(i + 20000)),
                                      ('Text', None)])} for i in range],
                            'GPX': [{'geometry': {'type': 'Point', 'coordinates': (0.0, float(i))},
                                     'properties': {'ele': 0.0, 'time': '2020-03-24T16:08:40+00:00'}} for i
                                    in range],
                            'GPSTrackMaker': [{'geometry': {'type': 'Point', 'coordinates': (0.0, float(i))},
                                               'properties': {}} for i in range],
                            'DGN': [
                                {'geometry': {'type': 'LineString', 'coordinates': [(float(i), 0.0), (0.0, 0.0)]},
                                 'properties': {}} for i in range],
                            'MapInfo File': [
                                {'geometry': {'type': 'Point', 'coordinates': (0.0, float(i))},
                                 'properties': {'position': str(i)}} for i in range],
                            'PCIDSK': [{'geometry': {'type': 'Point', 'coordinates': (0.0, float(i), 0.0)},
                                        'properties': {'position': i}} for i in range]
                            }
        return special_records1.get(driver, [
            {'geometry': {'type': 'Point', 'coordinates': (0.0, float(i))}, 'properties': {'position': i}} for i in
            range])

    def get_records2(driver, range):
        special_records2 = {'DGN': [
            {'geometry': {'type': 'LineString', 'coordinates': [(float(i), 0.0), (0.0, 0.0)]},
             'properties': OrderedDict(
                 [('Type', 4),
                  ('Level', 0),
                  ('GraphicGroup', 0),
                  ('ColorIndex', 0),
                  ('Weight', 0),
                  ('Style', 0),
                  ('EntityNum', None),
                  ('MSLink', None),
                  ('Text', None)])} for i in range],
        }
        return special_records2.get(driver, get_records(driver, range))

    def get_create_kwargs(driver):
        kwargs = {
            'FlatGeobuf': {'SPATIAL_INDEX': False}
        }
        return kwargs.get(driver, {})

    def test_equal(driver, val_in, val_out):
        is_good = True
        is_good = is_good and val_in['geometry'] == val_out['geometry']
        for key in val_in['properties']:
            if key in val_out['properties']:
                if driver == 'FileGDB' and isinstance(val_in['properties'][key], int):
                    is_good = is_good and str(val_in['properties'][key]) == str(int(val_out['properties'][key]))
                else:
                    is_good = is_good and str(val_in['properties'][key]) == str(val_out['properties'][key])
            else:
                is_good = False
        return is_good

    def _testdata_generator(driver, range1, range2):
        """ Generate test data and helper methods for a specific driver. Each set of generated set of records
        contains the position specified with range. These positions are either encoded as field or in the geometry
        of the record, depending of the driver characteristics.

        Parameters
        ----------
            driver: str
                Name of drive to generate tests for
            range1: list of integer
                Range of positions for first set of records
            range2: list of integer
                Range  of positions for second set of records

        Returns
        -------
        schema
            A schema for the records
        crs
            A crs for the records
        records1
            A set of records containing the positions of range1
        records2
            A set of records containing the positions of range2
        test_equal
            A function that returns True if the geometry is equal between the generated records and a record and if
            the properties of the generated records can be found in a record
        """
        return get_schema(driver), get_crs(driver), get_records(driver, range1), get_records2(driver, range2),\
               test_equal, get_create_kwargs(driver)

    return _testdata_generator


@pytest.fixture(scope='session')
def path_test_tz_geojson(data_dir):
    """Path to ```test_tz.geojson``"""
    return os.path.join(data_dir, 'test_tz.geojson')
